/*
  Copyright 2018-2021, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2021, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  This file is part of GekkoFS.

  GekkoFS is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GekkoFS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GekkoFS.  If not, see <https://www.gnu.org/licenses/>.

  SPDX-License-Identifier: GPL-3.0-or-later
*/

#include <filesystem>
#include "helpers/helpers.hpp"

#include <catch2/catch.hpp>
#include <fmt/format.h>
#include <daemon/backend/data/pmdk/pool.hpp>

namespace fs = std::filesystem;

using namespace std::string_literals;
using namespace Catch::Matchers;

// FIXME: temporary solution to have correct loggers for tests
#include <spdlog/sinks/basic_file_sink.h>

#include <vector>
#include <list>


namespace {

void
patch_gkfs_loggers(const std::vector<std::string>& loggers_name,
                   spdlog::level::level_enum level, const std::string& path) {

    /* Create common sink */
    auto file_sink = std::make_shared<spdlog::sinks::basic_file_sink_mt>(path);

    /* Create and configure loggers */
    auto loggers = std::list<std::shared_ptr<spdlog::logger>>();
    for(const auto& name : loggers_name) {

        if(spdlog::get(name)) {
            continue;
        }

        auto logger = std::make_shared<spdlog::logger>(name, file_sink);
        logger->flush_on(spdlog::level::trace);
        loggers.push_back(logger);
    }

    /* register loggers */
    for(const auto& logger : loggers) {
        spdlog::register_logger(logger);
    }

    // set logger format
    spdlog::set_pattern("[%C-%m-%d %H:%M:%S.%f] %P [%L][%n] %v");

    spdlog::set_level(level);
}

} // namespace

// end of temporary solution

SCENARIO("PMDK pools can be created", "[pmdk_storage_pool]") {


    auto logger_names = std::vector<std::string>{
            "main",
            "MetadataModule",
            "DataModule",
    };

    ::patch_gkfs_loggers(logger_names, spdlog::level::debug, "/dev/stdout");

    GIVEN("A valid parent directory ") {

        WHEN("Size is a positive integer") {

            std::size_t pool_size = 420;
            helpers::temporary_directory tmpdir;

            pmdk::pool p(tmpdir.dirname(), pool_size);

            THEN("The pool object is valid ") {
                REQUIRE(p);
                REQUIRE(p.valid());
                REQUIRE(p.size() == pool_size);
                REQUIRE(p.data() != nullptr);
            }

            AND_THEN("The mapping file exists and has the correct size") {
                REQUIRE(fs::exists(p.path()));
                REQUIRE(fs::file_size(p.path()) == pool_size);
            }
        }

        WHEN("Size is zero") {

            helpers::temporary_directory tmpdir;

            pmdk::pool p(tmpdir.dirname(), 0);

            THEN("The pool object is invalid ") {
                REQUIRE(!p);
                REQUIRE(!p.valid());

                AND_THEN("Internal members are default initialized") {
                    REQUIRE(p.path().empty());
                    REQUIRE(p.data() == nullptr);
                    REQUIRE(p.size() == 0);
                }
            }
        }

        WHEN("Size is negative") {

            helpers::temporary_directory tmpdir;

            pmdk::pool p(tmpdir.dirname(), -1);

            THEN("The pool object is invalid ") {
                REQUIRE(!p);
                REQUIRE(!p.valid());

                AND_THEN("Internal members are default initialized") {
                    REQUIRE(p.path().empty());
                    REQUIRE(p.data() == nullptr);
                    REQUIRE(p.size() == 0);
                }
            }
        }
    }

    GIVEN("An invalid parent directory ") {

        WHEN("Parent directory does not exist") {

            std::size_t pool_size = 420;
            helpers::temporary_directory tmpdir;

            pmdk::pool p("/foo/bar/", pool_size);

            THEN("The pool object is invalid ") {
                REQUIRE(!p);
                REQUIRE(!p.valid());

                AND_THEN("Internal members are default initialized") {
                    REQUIRE(p.path().empty());
                    REQUIRE(p.data() == nullptr);
                    REQUIRE(p.size() == 0);
                }
            }
        }
    }
}
